# -*- coding: utf-8 -*-
"""
Django settings for moo project.

Generated by 'django-admin startproject' using Django 4.2.5.

For more information on this file, see
https://docs.djangoproject.com/en/4.2/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/4.2/ref/settings/
"""

from pathlib import Path


# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve().parent.parent

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/4.2/howto/deployment/checklist/

ALLOWED_HOSTS = []


# Application definition
DEFAULT_GIT_REPO_URL = "https://gitlab.com/bubblehouse/django-moo.git"
DEFAULT_PERMISSIONS = (
    "anything",
    "read",
    "write",
    "entrust",
    "grant",
    "execute",
    "move",
    "transmute",
    "derive",
    "develop",
    "administer",
)

ALLOWED_MODULES = (
    "moo.core",
    "hashlib",
    "string",
    "re"
)

ALLOWED_BUILTINS = (
    "dict",
    "dir",
    "enumerate",
    "getattr",
    "hasattr",
    "list",
    "set",
    "sorted",
    "type",
)

# Here are all our supported prepositions
PREPOSITIONS = [
    ["with", "using"],
    ["at", "to"],
    ["in front of"],
    ["in", "inside", "into", "within"],
    ["on top of", "on", "onto", "upon", "above"],
    ["out of", "from inside", "from"],
    ["over"],
    ["through"],
    ["under", "underneath", "beneath", "below"],
    ["around", "round"],
    ["between", "among"],
    ["behind", "past"],
    ["beside", "by", "near", "next to", "along"],
    ["for", "about"],
    ['is'],
    ["as"],
    ["off", "off of"],
]

OBJECT_SPECIFIER_CHOICES = (("this", "this"), ("any", "any"), ("none", "none"), ("either", "either"))

PREPOSITION_SPECIFIER_CHOICES = (("any", "any"), ("none", "none"))

# When True, Parser.get_verb() issues a single batch query against AncestorCache
# instead of one _lookup_verb() call per search-order object.
# Requires AncestorCache to be migrated and populated (migration 0025).
# Set to False to fall back to the sequential dispatch path.
MOO_BATCH_VERB_DISPATCH = False

# TTL in seconds for cross-session property value caching in get_property().
# Set to 0 to disable (e.g. in test environments where the cache backend is
# in-process and does not reset between test cases).
MOO_PROP_CACHE_TTL = 120

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django_ace",
    "simplesshkey",
    "django_celery_results",
    "django_celery_beat",
    "moo.core",
    "moo.shell",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

STATICFILES_FINDERS = [
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
]

ROOT_URLCONF = "moo.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

WSGI_APPLICATION = "moo.wsgi.application"


# Database
# https://docs.djangoproject.com/en/4.2/ref/settings/#databases

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": BASE_DIR / "db.sqlite3",
    }
}


# Cache
# https://docs.djangoproject.com/en/4.2/ref/settings/#caches

CACHES = {
    "default": {
        "BACKEND": "django.core.cache.backends.redis.RedisCache",
        "LOCATION": "redis://localhost:6379",
    }
}

# Password validation
# https://docs.djangoproject.com/en/4.2/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]


# Internationalization
# https://docs.djangoproject.com/en/4.2/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/4.2/howto/static-files/

STATIC_URL = "admin/static/"

# Default primary key field type
# https://docs.djangoproject.com/en/4.2/ref/settings/#default-auto-field

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"

LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "filters": {
        "require_debug_false": {
            "()": "django.utils.log.RequireDebugFalse",
        },
        "require_debug_true": {
            "()": "django.utils.log.RequireDebugTrue",
        },
    },
    "formatters": {
        "simple": {
            "format": "%(asctime)s: %(levelname)s %(message)s",
        },
    },
    "handlers": {
        "console": {
            "level": "DEBUG",
            "class": "logging.StreamHandler",
            "formatter": "simple",
        },
    },
    "loggers": {
        "": {
            "handlers": ["console"],
            "level": "INFO",
            "propagate": False,
        }
    },
}



#!moo verb move --on $exit

# pylint: disable=return-outside-function,undefined-variable

"""
This verb is used to move `thing` through the exit. It provides a general mechanism for moving any sort of object
through the exit, not necessarily just players. The code for this verb performs a number of actions. First, the lock on
the exit is checked to see if thing is allowed to use the exit. If this is not the case, the `nogo_msg` and `onogo_msg`
text is sent to thing and everyone else in thing's location, respectively.

If the object is allowed to use the exit, it is blessed for entry to the destination room. This is done to ensure that
the object will be accepted by the destination room. It provides a way to stop objects moving into a room by any means
other than an exit leading into the room. By simply prohibiting all objects from entering the room, the only way in is
then to use an exit that leads into that room.

If the object is accepted by the room, determined using the $room:accept verb, then the leave messages are printed to
thing and the other people in the room. Then thing:moveto is invoked to move the object from the current room to the
destination of the exit. Once this has been done, the arrive messages for the exit are printed out to thing and the
destination room's occupants.
"""

from moo.core import context

thing = args[0]
source = this.get_property("source")
dest = this.get_property("dest")

if this.is_locked():
    thing.tell(this.nogo_msg(source, dest))
    source.announce_all_but(thing, this.onogo_msg(source, dest))
    return

dest.bless_for_entry(context.caller)
if dest.accept(thing):
    thing.tell(this.leave_msg(source, dest))
    source.announce_all_but(thing, this.oleave_msg(source, dest))
    thing.moveto(dest)
    thing.tell(this.arrive_msg(source, dest))
    dest.announce_all_but(thing, this.oarrive_msg(source, dest))
else:
    thing.tell(this.nogo_msg(source, dest))
    source.announce_all_but(thing, this.onogo_msg(source, dest))



#!moo verb open close unlock lock --on $exit --dspec any

# pylint: disable=return-outside-function,undefined-variable

from moo.core import context, PropertyDoesNotExist

door_description = context.parser.get_dobj_str()
door = context.caller.location.match_exit(door_description)
if not door:
    print(f"There is no door called {door_description} here.")
    return

# this is the simplest kind of door, where access control is
# determined by the ownership of the corresponding properties
if verb_name == "open":
    if door.is_open():
        print("The door is already open.")
    else:
        if door.is_locked():
            print("The door is locked.")
        else:
            door.set_property("open", True)
            print("The door is open.")
elif verb_name == "close":
    if not door.is_open():
        print("The door is already closed.")
    else:
        door.set_property("open", False)
        try:
            autolock = door.get_property("autolock")
        except PropertyDoesNotExist:
            autolock = False
        if autolock:
            door.set_property("locked", True)
        print("The door is closed.")
elif verb_name == "unlock":
    if door.is_locked():
        door.set_property("locked", False)
        print("The door is unlocked.")
    else:
        print("The door is not locked.")
elif verb_name == "lock":
    if door.is_locked():
        print("The door is already locked.")
    else:
        door.set_property("locked", True)
        print("The door is locked.")



# Generated by Django 4.2.5 on 2023-11-27 02:26

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("core", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="object",
            name="observers",
            field=models.ManyToManyField(
                blank=True, related_name="observations", through="core.Observation", to="core.object"
            ),
        ),
        migrations.AlterField(
            model_name="object",
            name="unique_name",
            field=models.BooleanField(default=False),
        ),
    ]



# -*- coding: utf-8 -*-
"""
Support resources for PyTest framework.
"""

import importlib.resources
import logging

import pytest
from django.conf import settings
from django.contrib.auth.models import User  # pylint: disable=imported-auth-user

from moo.bootstrap import load_python
from moo.core.models import Object, Player, Repository

log = logging.getLogger(__name__)


@pytest.fixture(autouse=True, scope="session")
def configure_celery_for_tests():
    from moo.celery import app
    app.conf.update(
        broker_url="memory://",
        task_always_eager=True,
        task_store_eager_result=True,
    )


@pytest.fixture()
def t_init(request):
    """
    Test fixture that pre-seeds a basic bootstrapped environment.
    """
    name = request.param if hasattr(request, "param") else "test"
    log.debug(f"t_init: {name}")
    Repository.objects.create(slug=name, prefix=f"moo/bootstrap/{name}_verbs", url=settings.DEFAULT_GIT_REPO_URL)
    ref = importlib.resources.files("moo.bootstrap") / f"{name}.py"
    with importlib.resources.as_file(ref) as path:
        load_python(path)
    yield Object.objects.get(id=1)


@pytest.fixture()
def t_wizard():
    """
    Test fixture that returns the Wizard account.
    """
    yield Object.objects.get(name="Wizard")



import importlib.resources

from django.conf import settings
from django.core.management.base import BaseCommand
from django.db import transaction

from moo.bootstrap import load_python
from moo.core.models import Repository

builtin_templates = ["minimal", "default"]


class Command(BaseCommand):
    help = "Initialize a new database."

    def add_arguments(self, parser):
        parser.add_argument(
            "--bootstrap",
            type=str,
            default="default",
            help="Optionally pass a built-in template name or a Python source file" " to bootstrap the database.",
        )

    def handle(self, *args, bootstrap="default", **config):
        with transaction.atomic():
            try:
                Repository.objects.get(slug=bootstrap)
                raise RuntimeError("Looks like this database has already been initialized.")
            except Repository.DoesNotExist:
                Repository.objects.create(
                    slug=bootstrap,
                    prefix=f"moo/bootstrap/{bootstrap}_verbs",
                    url=settings.DEFAULT_GIT_REPO_URL,
                )
            if bootstrap in builtin_templates:
                ref = importlib.resources.files("moo.bootstrap") / f"{bootstrap}.py"
                with importlib.resources.as_file(ref) as path:
                    bootstrap_path = path
            else:
                raise NotImplementedError(bootstrap)
            load_python(bootstrap_path)



#!moo verb @desc*ribe --on $player --dspec any --ispec as:any

# pylint: disable=return-outside-function,undefined-variable

"""
This is a player command used to set the description of an object. It takes the dobjstr and tries to match it with an
object. If a match is found, then the object's `describe` verb is invoked, with the iobjstr as an argument.
"""

from moo.core import context

if not (context.parser.has_dobj_str()):
    print("[red]What do you want to describe?[/red]")
    return
if not (context.parser.has_pobj_str("as")):
    print("[red]What do you want to describe that as?[/red]")
    return

subject = context.parser.get_dobj()
subject.describe(context.parser.get_pobj_str("as"))
print("[color yellow]Description set for %s[/color yellow]" % subject)



#!moo verb add_exit --on $room

# pylint: disable=return-outside-function,undefined-variable

"""
This verb is used to add exit to the list of exits leading out of the room. This verb, and the :match_exit verb provide
the interface to the room exits list. The way in which exits are stored, removed and matched has been separated from
the interface so that different implementations of the exits concept can be used in different sub classes of the $room
class.
"""

from moo.core import PropertyDoesNotExist

exit = args[0]

try:
    exits = this.get_property("exits")
except PropertyDoesNotExist:
    exits = []

exits.append(exit)
this.set_property("exits", exits)

return True



#!moo verb eject --on $root_class

# pylint: disable=return-outside-function,undefined-variable

"""
This verb is used to remove `something` from the contents of an object. The owner of an object, or a wizard, can use
this verb to eject a victim from inside the object. The victim is sent to `None`, for most objects, or to
`$player_start` if the victim is a player.
"""

something = args[0]
if this.contains(something):
    if something.is_player():
        something.moveto(_.player_start)
    else:
        something.moveto(None)
else:
    print(f"{this.name} does not contain {something.name}.")



# pylint: disable=undefined-variable

from moo.core import context, set_task_perms

obj = args[0]
with set_task_perms(context.player):
    obj.allow("wizards", "anything")
    obj.allow("owners", "anything")

    if obj.kind == "verb":
        obj.allow("everyone", "execute")
    elif obj.kind == "property":
        obj.allow("everyone", "read")
    elif obj.kind == "object":
        obj.allow("everyone", "read")



#!moo verb remove_exit --on $room

# pylint: disable=return-outside-function,undefined-variable

"""
This verb performs the opposite function to the `add_exit` verb. It removes `exit` from the room's list of exits. If it is
not possible to remove `exit` from the room's exit list (normally because the object that invoked the verb does not have
the required permission) then the verb returns `False`. Otherwise, a successful addition returns `True`.
"""

from moo.core import PropertyDoesNotExist

exit = args[0]

try:
    exits = this.get_property("exits")
except PropertyDoesNotExist:
    exits = []

exits.remove(exit)
this.set_property("exits", exits)

return True



#!moo verb accept --on $room

# pylint: disable=return-outside-function,undefined-variable

"""
This verb is used to control what objects are permitted to be placed inside other objects. If this verb returns
`False`, then the object where cannot be moved into this object. Conversely, if the verb returns a non-zero value, the
object is allowed to be placed inside this object. The verb is called by the server when it changes the `location`
property of an object.

The `$room` class definition provides for a flexible scheme using various different criteria, as shown in the following code:

       what = args[0]
       return this.is_unlocked_for(what) and (
              this.free_entry or
              what.owner == this.owner or
              what in this.residents)

Starting at the top of the conditional expression, we see the locking condition being checked. If the room lock
forbids this object to enter the room, then the :accept verb returns zero.

If this is not the case, then we consider the value of the `free_entry` property. If this is set to a non-zero
value, then the object is allowed to enter the room.

If the owner of an object is the owner of a room, the object is allowed to enter.

Finally, if a `residents` list is defined in the room, and the object is in the list, then it is allowed to enter.
This complex set of conditions shows how an arbitrary set of criteria can be applied to the movement of objects into other objects.
"""

what = args[0]
return this.is_unlocked_for(what) and (
       this.free_entry or
       what.owner == this.owner or
       what in this.residents)



import pytest

from moo.core import code, lookup, create
from moo.core.models import Object


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_parse_keyexp(t_init: Object, t_wizard: Object):
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.ContextManager(t_wizard, _writer):
        system = lookup(1)
        lock_utils = system.get_property("lock_utils")
        assert lock_utils is not None

        # Example 1: simple object number
        keyexp = "#123"
        key = lock_utils.parse_keyexp(keyexp)
        assert key == 123

        # Example 3: two objects
        keyexp = "#12 || #34"
        key = lock_utils.parse_keyexp(keyexp)
        assert key == ["||", 12, 34]

        # Example 4: negation
        keyexp = "!#56"
        key = lock_utils.parse_keyexp(keyexp)
        assert key == ["!", 56]

        # Example from docs:
        keyexp = "#45 && ?#46 && (#47 || !#48)"
        key = lock_utils.parse_keyexp(keyexp)
        assert key == ["&&", ["&&", 45, ["?", 46]], ["||", 47, ["!", 48]]]


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_eval_key_direct_match(t_init: Object, t_wizard: Object):
    """eval_key with an integer key returns True only for the matching object."""
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.ContextManager(t_wizard, _writer):
        system = lookup(1)
        lock_utils = system.get_property("lock_utils")
        assert lock_utils is not None

        obj = create("lock_target", parents=[system.thing])
        other = create("lock_other", parents=[system.thing])

        assert lock_utils.eval_key(obj.id, obj) is True
        assert lock_utils.eval_key(obj.id, other) is False


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_eval_key_not(t_init: Object, t_wizard: Object):
    """! negates the result; containment also satisfies the underlying check."""
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.ContextManager(t_wizard, _writer):
        system = lookup(1)
        lock_utils = system.get_property("lock_utils")
        assert lock_utils is not None

        container = create("lock_container", parents=[system.container])
        # item is inside container, so container.contains(item) is True
        item = create("lock_item", parents=[system.thing], location=container)

        # candidate is the item itself: not (item == item) → False
        assert lock_utils.eval_key(["!", item.id], item) is False
        # candidate is the container: not (container.contains(item)) → False
        assert lock_utils.eval_key(["!", item.id], container) is False
        # unrelated candidate: not False → True
        assert lock_utils.eval_key(["!", item.id], t_wizard) is True


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_eval_key_or(t_init: Object, t_wizard: Object):
    """|| returns True if either operand matches the candidate."""
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.ContextManager(t_wizard, _writer):
        system = lookup(1)
        lock_utils = system.get_property("lock_utils")
        assert lock_utils is not None

        obj_a = create("lock_or_a", parents=[system.thing])
        obj_b = create("lock_or_b", parents=[system.thing])
        obj_c = create("lock_or_c", parents=[system.thing])

        assert lock_utils.eval_key(["||", obj_a.id, obj_b.id], obj_a) is True
        assert lock_utils.eval_key(["||", obj_a.id, obj_b.id], obj_b) is True
        assert lock_utils.eval_key(["||", obj_a.id, obj_b.id], obj_c) is False


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_eval_key_and(t_init: Object, t_wizard: Object):
    """&& returns True only when the candidate contains both referenced objects."""
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.ContextManager(t_wizard, _writer):
        system = lookup(1)
        lock_utils = system.get_property("lock_utils")
        assert lock_utils is not None

        carrier = create("lock_carrier", parents=[system.container])
        thing1 = create("lock_thing1", parents=[system.thing], location=carrier)
        thing2 = create("lock_thing2", parents=[system.thing], location=carrier)

        # carrier holds both items
        assert lock_utils.eval_key(["&&", thing1.id, thing2.id], carrier) is True
        # t_wizard holds neither
        assert lock_utils.eval_key(["&&", thing1.id, thing2.id], t_wizard) is False


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_eval_key_question(t_init: Object, t_wizard: Object):
    """? evaluates the .key property of the referenced object against the candidate."""
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.ContextManager(t_wizard, _writer):
        system = lookup(1)
        lock_utils = system.get_property("lock_utils")
        assert lock_utils is not None

        key_holder = create("lock_key_holder", parents=[system.thing])
        candidate = create("lock_candidate", parents=[system.thing])

        # key_holder has no key property yet → False
        assert lock_utils.eval_key(["?", key_holder.id], candidate) is False

        # give key_holder a key that matches candidate
        key_holder.set_property("key", candidate.id)
        key_holder.save()
        assert lock_utils.eval_key(["?", key_holder.id], candidate) is True



#!moo verb look_self --on $player

# pylint: disable=return-outside-function,undefined-variable

"""
This verb overrides the `$root_class` definition to provide an indication to other players of whether this player is
currently active or not. It uses `passthrough()` to allow the parent class to print a description, and then looks at
the `connected_players()` list to determine if this player is currently connected. If not, then the text

    He is sleeping

is printed. If the player is connected, and has been idle for less than 60 seconds, then the string

    He is awake and looks alert

is printed. If the player is connected, but has been inactive for more than 60 seconds, the string

    He is awake, but has been staring off into space for X

is printed, where X is an indication of the time the player has been inactive. The gender pronoun inserted is taken
from the pronoun list for the player. This means it can vary with the gender of the player object.

If the player is carrying any objects, a simple list of these is printed.
"""

passthrough()
print(_.string_utils.pronoun_sub("%S is sleeping."))

# # TODO: implement this
# if this not in connected_players():
#     print("He is sleeping.")
# elif this.idle_time() < 60:
#     print("He is awake and looks alert.")
# else:
#     print(f"He is awake, but has been staring off into space for {this.idle_time()} seconds.")



# Generated by Django 5.1.4 on 2025-05-03 19:10

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0013_verb_direct_object_alter_prepositionname_preposition_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='prepositionname',
            name='preposition',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='names', to='core.preposition'),
        ),
    ]



from moo.core import parse


def test_lex_imperative_command():
    lex = parse.Lexer("look")
    assert lex.command == "look"


def test_lex_direct_object():
    lex = parse.Lexer("look here")
    assert lex.dobj_str == "here"


def test_lex_object_of_the_preposition():
    lex = parse.Lexer("look at this")
    assert lex.prepositions["at"][0][1] == "this"


def test_lex_direct_object_with_preposition():
    lex = parse.Lexer("look at painting with the glasses")
    assert lex.prepositions["at"][0][1] == "painting"
    assert lex.prepositions["with"][0][0] == "the"
    assert lex.prepositions["with"][0][1] == "glasses"


def test_lex_look_at_QUOTED_painting_with_the_glasses():
    lex = parse.Lexer("look at 'painting with the glasses'")
    assert lex.prepositions["at"][0][1] == "painting with the glasses"



# Generated by Django 5.1.4 on 2026-02-21 18:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0021_alter_verb_direct_object'),
    ]

    operations = [
        migrations.DeleteModel(
            name='AccessibleObject',
        ),
        migrations.DeleteModel(
            name='AccessibleProperty',
        ),
        migrations.DeleteModel(
            name='AccessibleVerb',
        ),
        migrations.RenameField(
            model_name='property',
            old_name='inherited',
            new_name='inherit_owner',
        ),
        migrations.AlterField(
            model_name='prepositionspecifier',
            name='specifier',
            field=models.CharField(choices=[('this', 'this'), ('any', 'any'), ('none', 'none'), ('either', 'either')], db_index=True, default='none', max_length=255),
        ),
        migrations.AlterField(
            model_name='verb',
            name='direct_object',
            field=models.CharField(choices=[('this', 'this'), ('any', 'any'), ('none', 'none'), ('either', 'either')], db_default='none', db_index=True, default='none', max_length=255),
        ),
    ]



import pytest

from moo.core import context, code, create, exceptions, lookup, parse
from moo.core.models import Object, Player


def setup_room(t_wizard: Object, name: str = "Test Room", description: str = "A plain test room."):
    """Create a Generic Room, describe it, and move the wizard into it."""
    rooms = lookup("Generic Room")
    room = create(name, parents=[rooms])
    room.describe(description)
    t_wizard.location = room
    t_wizard.save()
    context.caller.refresh_from_db()
    return room


def setup_item(location: Object, name: str = "red ball"):
    """Create a plain root_class child object in the given location."""
    system = lookup(1)
    return create(name, parents=[system.root_class], location=location)


# --- look_self ---


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_look_self_prints_room_name(t_init: Object, t_wizard: Object):
    """look_self() prints the room name in bright_yellow."""
    printed = []
    with code.ContextManager(t_wizard, printed.append):
        room = setup_room(t_wizard)
        room.look_self()
    assert f"[color bright_yellow]{room.name}[/color bright_yellow]" in printed


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_look_self_prints_description(t_init: Object, t_wizard: Object):
    """look_self() prints the room description via passthrough."""
    printed = []
    with code.ContextManager(t_wizard, printed.append):
        room = setup_room(t_wizard, description="A damp cave draped in shadows.")
        room.look_self()
    assert room.description() in printed


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_look_self_includes_contents(t_init: Object, t_wizard: Object):
    """look_self() calls tell_contents so room items appear in output."""
    printed = []
    with code.ContextManager(t_wizard, printed.append):
        room = setup_room(t_wizard)
        setup_item(room, "gold coin")
        room.look_self()
    assert any("gold coin" in line for line in printed)


# --- tell_contents: ctype 3 (default) ---


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_tell_contents_ctype3_empty(t_init: Object, t_wizard: Object):
    """tell_contents() with ctype=3 prints nothing when the room is empty."""
    printed = []
    with code.ContextManager(t_wizard, printed.append):
        rooms = lookup("Generic Room")
        room = create("Empty Room", parents=[rooms])
        room.set_property("content_list_type", 3)
        room.tell_contents()
    assert not printed


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_tell_contents_ctype3_single_item(t_init: Object, t_wizard: Object):
    """tell_contents() with ctype=3 prints a single non-player item."""
    printed = []
    with code.ContextManager(t_wizard, printed.append):
        rooms = lookup("Generic Room")
        room = create("One Item Room", parents=[rooms])
        room.set_property("content_list_type", 3)
        setup_item(room, "red ball")
        room.tell_contents()
    assert printed == ["You see red ball here."]


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_tell_contents_ctype3_multiple_items(t_init: Object, t_wizard: Object):
    """tell_contents() with ctype=3 lists multiple non-player items."""
    printed = []
    with code.ContextManager(t_wizard, printed.append):
        rooms = lookup("Generic Room")
        room = create("Multi Item Room", parents=[rooms])
        room.set_property("content_list_type", 3)
        setup_item(room, "red ball")
        setup_item(room, "blue box")
        room.tell_contents()
    assert printed == ["You see red ball and blue box here."]


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_tell_contents_ctype3_single_player(t_init: Object, t_wizard: Object):
    """tell_contents() with ctype=3 prints a single player."""
    printed = []
    with code.ContextManager(t_wizard, printed.append):
        rooms = lookup("Generic Room")
        room = create("Player Room", parents=[rooms])
        room.set_property("content_list_type", 3)
        t_wizard.location = room
        t_wizard.save()
        context.caller.refresh_from_db()
        player = lookup("Player")
        player.location = room
        player.save()
        room.tell_contents()
    assert printed == ["You see Player here."]


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_tell_contents_ctype3_multiple_players(t_init: Object, t_wizard: Object):
    """tell_contents() with ctype=3 lists multiple players with 'are here'."""
    printed = []
    with code.ContextManager(t_wizard, printed.append):
        rooms = lookup("Generic Room")
        room = create("Two Player Room", parents=[rooms])
        room.set_property("content_list_type", 3)
        t_wizard.location = room
        t_wizard.save()
        context.caller.refresh_from_db()
        player = lookup("Player")
        player.location = room
        player.save()
        bob = create("Bob", location=room)
        Player.objects.create(avatar=bob)
        room.tell_contents()
    assert printed == ["Player and Bob are here."]


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_tell_contents_ctype3_mixed(t_init: Object, t_wizard: Object):
    """tell_contents() with ctype=3 prints items and players on separate lines."""
    printed = []
    with code.ContextManager(t_wizard, printed.append):
        rooms = lookup("Generic Room")
        room = create("Mixed Room", parents=[rooms])
        room.set_property("content_list_type", 3)
        t_wizard.location = room
        t_wizard.save()
        context.caller.refresh_from_db()
        setup_item(room, "red ball")
        player = lookup("Player")
        player.location = room
        player.save()
        room.tell_contents()
    assert printed == ["You see red ball here.", "You see Player here."]


# --- tell_contents: ctype 2 ---


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_tell_contents_ctype2_single_item(t_init: Object, t_wizard: Object):
    """tell_contents() with ctype=2 prints a single item."""
    printed = []
    with code.ContextManager(t_wizard, printed.append):
        rooms = lookup("Generic Room")
        room = create("Ctype2 Room", parents=[rooms])
        room.set_property("content_list_type", 2)
        setup_item(room, "red ball")
        room.tell_contents()
    assert printed == ["You see red ball here."]


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_tell_contents_ctype2_mixed(t_init: Object, t_wizard: Object):
    """tell_contents() with ctype=2 lists players and items together."""
    printed = []
    with code.ContextManager(t_wizard, printed.append):
        rooms = lookup("Generic Room")
        room = create("Ctype2 Mixed Room", parents=[rooms])
        room.set_property("content_list_type", 2)
        t_wizard.location = room
        t_wizard.save()
        context.caller.refresh_from_db()
        player = lookup("Player")
        player.location = room
        player.save()
        setup_item(room, "red ball")
        room.tell_contents()
    assert printed == ["You see Player and red ball here."]


# --- tell_contents: ctype 1 ---


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_tell_contents_ctype1(t_init: Object, t_wizard: Object):
    """tell_contents() with ctype=1 prints one line per object."""
    printed = []
    with code.ContextManager(t_wizard, printed.append):
        rooms = lookup("Generic Room")
        room = create("Ctype1 Room", parents=[rooms])
        room.set_property("content_list_type", 1)
        t_wizard.location = room
        t_wizard.save()
        context.caller.refresh_from_db()
        player = lookup("Player")
        player.location = room
        player.save()
        setup_item(room, "red ball")
        room.tell_contents()
    assert printed == ["Player is here", "You see red ball here"]


# --- tell_contents: ctype 0 ---


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_tell_contents_ctype0(t_init: Object, t_wizard: Object):
    """tell_contents() with ctype=0 prints a raw Contents: list."""
    printed = []
    with code.ContextManager(t_wizard, printed.append):
        rooms = lookup("Generic Room")
        room = create("Ctype0 Room", parents=[rooms])
        room.set_property("content_list_type", 0)
        t_wizard.location = room
        t_wizard.save()
        context.caller.refresh_from_db()
        setup_item(room, "red ball")
        room.tell_contents()
    assert printed == ["Contents:", "red ball"]


# --- tell_contents: out-of-range ctype ---


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_tell_contents_ctype_out_of_range(t_init: Object, t_wizard: Object):
    """tell_contents() with an unrecognised ctype prints nothing."""
    printed = []
    with code.ContextManager(t_wizard, printed.append):
        rooms = lookup("Generic Room")
        room = create("Silent Room", parents=[rooms])
        room.set_property("content_list_type", 99)
        setup_item(room, "red ball")
        room.tell_contents()
    assert not printed


# --- confunc ---


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_confunc_shows_room_to_player(t_init: Object, t_wizard: Object):
    """confunc() displays the room's look_self output to the connecting player."""
    printed = []
    with code.ContextManager(t_wizard, printed.append):
        room = setup_room(t_wizard)
        room.confunc()
    assert f"[color bright_yellow]{room.name}[/color bright_yellow]" in printed


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_confunc_announces_connection_to_others(t_init: Object, t_wizard: Object):
    """confunc() tells everyone else in the room that the player has connected."""
    printed = []
    with code.ContextManager(t_wizard, printed.append):
        room = setup_room(t_wizard)
        player = lookup("Player")
        player.location = room
        player.save()
        with pytest.warns(RuntimeWarning, match=r"ConnectionError") as warnings:
            room.confunc()
    messages = [str(w.message) for w in warnings.list]
    assert any("has connected" in m and str(t_wizard) in m for m in messages)


# --- disfunc ---


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_disfunc_moves_player_to_home(t_init: Object, t_wizard: Object):
    """disfunc() teleports the player to their home when they are elsewhere."""
    with code.ContextManager(t_wizard, lambda msg: None):
        rooms = lookup("Generic Room")
        away = create("Away Room", parents=[rooms])
        t_player = lookup("Player")
        t_player.location = away
        t_player.save()
        t_wizard.location = away
        t_wizard.save()
        context.caller.refresh_from_db()
        with pytest.warns(RuntimeWarning, match=r"ConnectionError") as warnings:
            t_wizard.location.disfunc()
        t_wizard.refresh_from_db()
    messages = [str(w.message) for w in warnings.list]
    assert any(f"(Player)): #{t_wizard.pk} (Wizard) has disconnected." in m for m in messages)
    system = lookup(1)
    assert t_wizard.location == system.player_start


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_disfunc_noop_when_already_home(t_init: Object, t_wizard: Object):
    """disfunc() does nothing when the player is already at their home."""
    with code.ContextManager(t_wizard, lambda msg: None):
        system = lookup(1)
        t_wizard.location = system.player_start
        t_wizard.save()
        context.caller.refresh_from_db()
        home_before = system.player_start
        with pytest.warns(RuntimeWarning, match=r"ConnectionError") as warnings:
            t_wizard.location.disfunc()
        t_wizard.refresh_from_db()
    messages = [str(w.message) for w in warnings.list]
    assert any(f"(Player)): #{t_wizard.pk} (Wizard) has disconnected." in m for m in messages)
    assert t_wizard.location == home_before


# --- look (player command) ---


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_look_no_args_shows_current_room(t_init: Object, t_wizard: Object):
    """'look' with no arguments shows the current room's look_self."""
    printed = []
    with code.ContextManager(t_wizard, printed.append) as ctx:
        room = setup_room(t_wizard)
        parse.interpret(ctx, "look")
    assert f"[color bright_yellow]{room.name}[/color bright_yellow]" in printed


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_look_at_named_object(t_init: Object, t_wizard: Object):
    """'look <name>' shows the description of the named object in the room."""
    printed = []
    with code.ContextManager(t_wizard, printed.append) as ctx:
        room = setup_room(t_wizard)
        coin = setup_item(room, "gold coin")
        coin.describe("A shiny gold coin minted in ancient times.")
        parse.interpret(ctx, "look gold coin")
    assert any("ancient times" in line for line in printed)


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_look_missing_object_prints_error(t_init: Object, t_wizard: Object):
    """'look <name>' prints an error when no matching object is present."""
    printed = []
    with code.ContextManager(t_wizard, printed.append) as ctx:
        setup_room(t_wizard)
        parse.interpret(ctx, "look nonexistent thing")
    assert any("nonexistent thing" in line for line in printed)


# --- say ---


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_say_delivers_to_caller_and_others(t_init: Object, t_wizard: Object):
    """say sends 'You: msg' to the caller and 'Name: msg' to others in the room."""
    with code.ContextManager(t_wizard, lambda msg: None) as ctx:
        with pytest.warns(RuntimeWarning, match=r"ConnectionError") as warnings:
            parse.interpret(ctx, "say Hello there!")
    messages = [str(w.message) for w in warnings.list]
    assert any("(Wizard)): You: Hello there!" in m for m in messages)
    assert any("(Player)): Wizard: Hello there!" in m for m in messages)


# --- emote ---


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_emote_caller_sees_own_action(t_init: Object, t_wizard: Object):
    """emote sends 'You <action>' to the caller."""
    with code.ContextManager(t_wizard, lambda msg: None) as ctx:
        with pytest.warns(RuntimeWarning, match=r"ConnectionError") as warnings:
            parse.interpret(ctx, "emote waves hello.")
    messages = [str(w.message) for w in warnings.list]
    assert any("(Wizard)): You waves hello." in m for m in messages)


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_emote_others_see_action(t_init: Object, t_wizard: Object):
    """emote sends the action text to others in the room via announce."""
    with code.ContextManager(t_wizard, lambda msg: None) as ctx:
        with pytest.warns(RuntimeWarning, match=r"ConnectionError") as warnings:
            parse.interpret(ctx, "emote waves hello.")
    messages = [str(w.message) for w in warnings.list]
    assert any("(Player)): waves hello." in m for m in messages)


# --- announce ---


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_announce_skips_caller(t_init: Object, t_wizard: Object):
    """announce() does not deliver the message to the caller."""
    received_by_wizard = []
    with code.ContextManager(t_wizard, received_by_wizard.append):
        room = setup_room(t_wizard)
        room.announce("secret message")
    assert not any("secret message" in line for line in received_by_wizard)


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_announce_delivers_to_others(t_init: Object, t_wizard: Object):
    """announce() sends the message to every room occupant except the caller."""
    with code.ContextManager(t_wizard, lambda msg: None):
        room = setup_room(t_wizard)
        player = lookup("Player")
        player.location = room
        player.save()
        with pytest.warns(RuntimeWarning, match=r"ConnectionError") as warnings:
            room.announce("broadcast message")
    messages = [str(w.message) for w in warnings.list]
    assert any("(Player)): broadcast message" in m for m in messages)
    assert not any("(Wizard)): broadcast message" in m for m in messages)


# --- announce_all ---


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_announce_all_delivers_to_everyone(t_init: Object, t_wizard: Object):
    """announce_all() sends the message to every occupant including the caller."""
    with code.ContextManager(t_wizard, lambda msg: None):
        room = setup_room(t_wizard)
        player = lookup("Player")
        player.location = room
        player.save()
        with pytest.warns(RuntimeWarning, match=r"ConnectionError") as warnings:
            room.announce_all("all hands message")
    messages = [str(w.message) for w in warnings.list]
    assert any("(Wizard)): all hands message" in m for m in messages)
    assert any("(Player)): all hands message" in m for m in messages)


# --- announce_all_but ---


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_announce_all_but_skips_specified_object(t_init: Object, t_wizard: Object):
    """announce_all_but() skips exactly the specified object."""
    with code.ContextManager(t_wizard, lambda msg: None):
        room = setup_room(t_wizard)
        player = lookup("Player")
        player.location = room
        player.save()
        with pytest.warns(RuntimeWarning, match=r"ConnectionError") as warnings:
            room.announce_all_but(player, "exclusive message")
    messages = [str(w.message) for w in warnings.list]
    assert any("(Wizard)): exclusive message" in m for m in messages)
    assert not any("(Player)): exclusive message" in m for m in messages)


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_announce_all_but_delivers_to_rest(t_init: Object, t_wizard: Object):
    """announce_all_but() delivers the message to all non-excluded occupants."""
    with code.ContextManager(t_wizard, lambda msg: None):
        room = setup_room(t_wizard)
        player = lookup("Player")
        player.location = room
        player.save()
        with pytest.warns(RuntimeWarning, match=r"ConnectionError") as warnings:
            room.announce_all_but(t_wizard, "player only message")
    messages = [str(w.message) for w in warnings.list]
    assert any("(Player)): player only message" in m for m in messages)
    assert not any("(Wizard)): player only message" in m for m in messages)


# --- huh / huh2 ---


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_huh_unrecognised_command(t_init: Object, t_wizard: Object):
    """An unrecognised command triggers huh2, which tells the player 'Huh?'."""
    with code.ContextManager(t_wizard, lambda msg: None) as ctx:
        with pytest.warns(RuntimeWarning, match=r"ConnectionError") as warnings:
            parse.interpret(ctx, "xyzzy")
    messages = [str(w.message) for w in warnings.list]
    assert any("Huh?" in m for m in messages)


# --- accept ---


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_accept_free_entry_allows_any_object(t_init: Object, t_wizard: Object):
    """accept() returns True when the room's free_entry property is set."""
    with code.ContextManager(t_wizard, lambda msg: None):
        rooms = lookup("Generic Room")
        room = create("Open Room", parents=[rooms])
        room.set_property("free_entry", True)
        item = setup_item(t_wizard.location)
        assert room.accept(item) is True


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_accept_owner_match_allows_entry(t_init: Object, t_wizard: Object):
    """accept() returns True when the object's owner matches the room's owner."""
    with code.ContextManager(t_wizard, lambda msg: None):
        rooms = lookup("Generic Room")
        room = create("Owned Room", parents=[rooms])
        room.set_property("free_entry", False)
        # wizard owns both the room and the item by default
        item = setup_item(t_wizard.location)
        assert room.accept(item) is True


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_accept_locked_room_denies_entry(t_init: Object, t_wizard: Object):
    """accept() returns False when the room is locked against the object."""
    with code.ContextManager(t_wizard, lambda msg: None):
        rooms = lookup("Generic Room")
        room = create("Locked Room", parents=[rooms])
        room.set_property("free_entry", False)
        item = setup_item(t_wizard.location)
        # Point the key at the room itself — the item won't satisfy that lock
        room.set_property("key", room)
        assert room.accept(item) is False


# --- match_exit ---


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_match_exit_found_by_name(t_init: Object, t_wizard: Object):
    """match_exit() returns the exit object when matched by direction name."""
    printed = []
    with code.ContextManager(t_wizard, printed.append) as ctx:
        parse.interpret(ctx, "@dig north to Destination Room")
        exit_obj = t_wizard.location.match_exit("north")
    assert exit_obj is not None
    assert exit_obj.dest.name == "Destination Room"


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_match_exit_not_found_returns_none(t_init: Object, t_wizard: Object):
    """match_exit() returns None when no exit matches the given direction."""
    with code.ContextManager(t_wizard, lambda msg: None):
        result = t_wizard.location.match_exit("south")
    assert result is None


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_match_exit_case_insensitive(t_init: Object, t_wizard: Object):
    """match_exit() matches direction names case-insensitively."""
    printed = []
    with code.ContextManager(t_wizard, printed.append) as ctx:
        parse.interpret(ctx, "@dig east to East Wing")
        exit_obj = t_wizard.location.match_exit("EAST")
    assert exit_obj is not None


# --- match_object ---


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_match_object_found_in_room(t_init: Object, t_wizard: Object):
    """match_object() returns the object when it is in the room."""
    with code.ContextManager(t_wizard, lambda msg: None):
        room = setup_room(t_wizard)
        coin = setup_item(room, "silver coin")
        result = room.match_object("silver coin")
    assert result == coin


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_match_object_not_found_raises(t_init: Object, t_wizard: Object):
    """match_object() raises DoesNotExist when nothing matches the name."""
    with code.ContextManager(t_wizard, lambda msg: None):
        room = setup_room(t_wizard)
        with pytest.raises(Object.DoesNotExist):
            room.match_object("invisible thing")


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_match_object_ambiguous_raises(t_init: Object, t_wizard: Object):
    """match_object() raises AmbiguousObjectError when multiple objects share a name."""
    with code.ContextManager(t_wizard, lambda msg: None):
        room = setup_room(t_wizard)
        setup_item(room, "red key")
        setup_item(room, "red key")
        with pytest.raises(exceptions.AmbiguousObjectError):
            room.match_object("red key")


# --- @exits ---


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_at_exits_no_exits(t_init: Object, t_wizard: Object):
    """@exits prints an error message when the room has no exits."""
    printed = []
    with code.ContextManager(t_wizard, printed.append) as ctx:
        setup_room(t_wizard)
        parse.interpret(ctx, "@exits")
    assert printed == ["[color red]There are no exits defined for this room.[/color red]"]


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_at_exits_lists_exits(t_init: Object, t_wizard: Object):
    """@exits lists each exit with its aliases and destination."""
    printed = []
    with code.ContextManager(t_wizard, printed.append) as ctx:
        setup_room(t_wizard)
        parse.interpret(ctx, "@dig north to The North Hall")
        printed.clear()
        parse.interpret(ctx, "@exits")
    assert printed[0] == "[color cyan]Exits defined for this room:[/color cyan]"
    assert any("north" in line and "The North Hall" in line for line in printed)


# --- @entrances ---


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_at_entrances_no_entrances(t_init: Object, t_wizard: Object):
    """@entrances prints an error message when the room has no entrances."""
    printed = []
    with code.ContextManager(t_wizard, printed.append) as ctx:
        setup_room(t_wizard)
        parse.interpret(ctx, "@entrances")
    assert printed == ["[color red]There are no entrances defined for this room.[/color red]"]


@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_at_entrances_lists_entrances(t_init: Object, t_wizard: Object):
    """@entrances lists each entrance with its aliases and source room."""
    printed = []
    with code.ContextManager(t_wizard, printed.append) as ctx:
        home = t_wizard.location
        setup_room(t_wizard, name="Far Room")
        parse.interpret(ctx, f"@tunnel south to {home.name}")
        # @tunnel added the entrance to `home`, so move back there before querying
        t_wizard.location = home
        t_wizard.save()
        context.caller.refresh_from_db()
        printed.clear()
        parse.interpret(ctx, "@entrances")
    assert printed[0] == "[color cyan]Entrances defined for this room:[/color cyan]"
    assert any("south" in line and home.name in line for line in printed)

@pytest.mark.django_db(transaction=True, reset_sequences=True)
@pytest.mark.parametrize("t_init", ["default"], indirect=True)
def test_basic_dig_and_tunnel(t_init: Object, t_wizard: Object):
    """dig and tunnel integration"""
    printed = []

    def _writer(msg):
        printed.append(msg)

    t_player = lookup("Player")
    with code.ContextManager(t_wizard, _writer) as ctx:
        home_location = t_wizard.location
        parse.interpret(ctx, "@dig north to Another Room")
        another_room = lookup("Another Room")
        assert printed == [
            '[color yellow]Dug an exit north to "Another Room".[/color yellow]',
        ]
        printed.clear()

        parse.interpret(ctx, "@dig north to Another Room")
        assert printed == ["[color red]There is already an exit in that direction.[/color red]"]
        printed.clear()

    with code.ContextManager(t_player, _writer) as ctx:
        with pytest.warns(RuntimeWarning, match=r"ConnectionError") as warnings:
            parse.interpret(ctx, "go north")
        assert [str(x.message) for x in warnings.list] == [
            f"ConnectionError(#{t_player.pk} (Player)): You leave #{home_location.pk} (The Laboratory).",
            f"ConnectionError(#{t_wizard.pk} (Wizard)): #{t_player.pk} (Player) leaves #{home_location.pk} (The Laboratory).",
            f"ConnectionError(#{t_player.pk} (Player)): You arrive at #{another_room.pk} (Another Room)."
        ]
        t_player.refresh_from_db()
        assert t_player.location.name == "Another Room"
        printed.clear()

    with code.ContextManager(t_wizard, _writer) as ctx:
        with pytest.warns(RuntimeWarning, match=r"ConnectionError") as warnings:
            parse.interpret(ctx, "go north")
        assert [str(x.message) for x in warnings.list] == [
            f"ConnectionError(#{t_wizard.pk} (Wizard)): You leave #{home_location.pk} (The Laboratory).",
            f"ConnectionError(#{t_wizard.pk} (Wizard)): You arrive at #{another_room.pk} (Another Room).",
            f"ConnectionError(#{t_player.pk} (Player)): #{t_wizard.pk} (Wizard) arrives at #{another_room.pk} (Another Room)."
        ]
        context.caller.refresh_from_db()
        context.player.refresh_from_db()
        parse.interpret(ctx, f"@tunnel south to {home_location.name}")
        assert printed == [
            f'[color yellow]Tunnelled an exit south to "{home_location.name}".[/color yellow]',
        ]
        assert t_player.location.get_property('exits')
        printed.clear()

    with code.ContextManager(t_player, _writer) as ctx:
        with pytest.warns(RuntimeWarning, match=r"ConnectionError") as warnings:
            parse.interpret(ctx, "go south")
        assert [str(x.message) for x in warnings.list] == [
            f"ConnectionError(#{t_player.pk} (Player)): You leave #{another_room.pk} (Another Room).",
            f"ConnectionError(#{t_wizard.pk} (Wizard)): #{t_player.pk} (Player) leaves #{another_room.pk} (Another Room).",
            f"ConnectionError(#{t_player.pk} (Player)): You arrive at #{home_location.pk} (The Laboratory)."
        ]
        t_player.refresh_from_db()
        assert t_player.location.name == home_location.name
        printed.clear()

    with code.ContextManager(t_player, _writer) as ctx:
        parse.interpret(ctx, "@exits")
        assert printed == [
            "[color cyan]Exits defined for this room:[/color cyan]",
            f"- [color yellow]north from The Laboratory[/color yellow] (Aliases: north) to [color green]Another Room[/color green] (#{another_room.pk})"
        ]
        printed.clear()

    with code.ContextManager(t_player, _writer) as ctx:
        parse.interpret(ctx, "@entrances")
        assert printed == [
            "[color cyan]Entrances defined for this room:[/color cyan]",
            f"- [color yellow]south from Another Room[/color yellow] (Aliases: south) to [color green]The Laboratory[/color green] (#{home_location.pk})"
        ]
        printed.clear()



import logging

import pytest
from django.db import connection
from django.test import override_settings

from moo.core.models import Object, Player

from .. import code, create, exceptions, lookup


@pytest.mark.django_db(transaction=True, reset_sequences=True)
def test_regular_user_can_read_a_thing(t_init: Object, t_wizard: Object):
    thing = Object.objects.create(name="thing", owner=t_wizard)
    user = Object.objects.get(name__iexact="player")
    assert user.is_allowed("read", thing)
    assert not user.is_allowed("write", thing)
    assert not user.is_allowed("entrust", thing)
    assert not user.is_allowed("move", thing)
    assert not user.is_allowed("transmute", thing)
    assert not user.is_allowed("derive", thing)
    assert not user.is_allowed("develop", thing)


@pytest.mark.django_db(transaction=True, reset_sequences=True)
def test_regular_user_who_owns_a_thing(t_init: Object, t_wizard: Object):
    user = Object.objects.get(name__iexact="player")
    thing = Object.objects.create(name="thing", owner=user)
    assert user.is_allowed("read", thing)
    assert user.is_allowed("write", thing)
    assert user.is_allowed("entrust", thing)
    assert user.is_allowed("move", thing)
    assert user.is_allowed("transmute", thing)
    assert user.is_allowed("derive", thing)
    assert user.is_allowed("develop", thing)


@pytest.mark.django_db(transaction=True, reset_sequences=True)
def test_everyone_can_read_a_thing(t_init: Object, t_wizard: Object):
    thing = Object.objects.create(name="thing")
    jim = Object.objects.create(name="Jim", unique_name=True)
    assert jim.is_allowed("read", thing)
    assert not jim.is_allowed("write", thing)
    assert not jim.is_allowed("entrust", thing)
    assert not jim.is_allowed("move", thing)
    assert not jim.is_allowed("transmute", thing)
    assert not jim.is_allowed("derive", thing)
    assert not jim.is_allowed("develop", thing)


@pytest.mark.django_db(transaction=True, reset_sequences=True)
def test_wizard_can_do_most_things(t_init: Object, t_wizard: Object):
    # Objects are only Wizards if they have an associated Player with wizard=True
    Player.objects.create(avatar=t_wizard, wizard=True)
    thing = Object.objects.create(name="thing")
    assert t_wizard.is_allowed("read", thing)
    assert t_wizard.is_allowed("write", thing)
    assert t_wizard.is_allowed("entrust", thing)
    assert t_wizard.is_allowed("move", thing)
    assert t_wizard.is_allowed("transmute", thing)
    assert t_wizard.is_allowed("derive", thing)
    assert t_wizard.is_allowed("develop", thing)


@pytest.mark.django_db(transaction=True, reset_sequences=True)
def test_add_a_simple_deny_clase(t_init: Object, t_wizard: Object):
    user = Object.objects.get(name__iexact="player")
    thing = Object.objects.create(name="thing", owner=user)
    thing.allow("everyone", "anything")
    thing.deny(user, "write")

    assert user.is_allowed("read", thing)
    assert not user.is_allowed("write", thing)


@pytest.mark.django_db(transaction=True, reset_sequences=True)
def test_cant_create_child_of_an_object_that_isnt_yours(t_init: Object, t_wizard: Object):
    printed = []

    def _writer(msg):
        printed.append(msg)

    user = Object.objects.get(name__iexact="player")
    parent_thing = Object.objects.create(name="parent thing", owner=t_wizard)
    with code.ContextManager(user, _writer):
        child_thing = Object.objects.create(name="child thing", owner=user)
        with pytest.raises(PermissionError) as excinfo:
            child_thing.parents.add(parent_thing)
        assert str(excinfo.value) == f"#{user.pk} (Player) is not allowed derive on #{parent_thing.pk} (parent thing)"


@pytest.mark.django_db(transaction=True, reset_sequences=True)
def test_cant_create_parent_of_an_object_that_isnt_yours(t_init: Object, t_wizard: Object):
    printed = []

    def _writer(msg):
        printed.append(msg)

    user = Object.objects.get(name__iexact="player")
    child_thing = Object.objects.create(name="child thing", owner=t_wizard)
    with code.ContextManager(user, _writer):
        parent_thing = Object.objects.create(name="parent thing", owner=user)
        with pytest.raises(PermissionError) as excinfo:
            child_thing.parents.add(parent_thing)
        assert str(excinfo.value) == f"#{user.pk} (Player) is not allowed transmute on #{child_thing.pk} (child thing)"


@pytest.mark.django_db(transaction=True, reset_sequences=True)
def test_cant_change_owner_unless_allowed_to_entrust(t_init: Object, t_wizard: Object):
    printed = []

    def _writer(msg):
        printed.append(msg)

    user = Object.objects.get(name__iexact="player")
    with code.ContextManager(user, _writer):
        with pytest.raises(PermissionError) as excinfo:
            create("thing", owner=t_wizard)
        assert str(excinfo.value) == "Can't change owner at creation time."
        obj = create("thing")
        with pytest.raises(PermissionError) as excinfo:
            obj.owner = t_wizard
            obj.save()
        assert str(excinfo.value) == f"#{user.pk} (Player) is not allowed entrust on #{obj.pk} (thing)"
    with code.ContextManager(t_wizard, _writer):
        obj = lookup("thing")
        obj.allow(user, "entrust")
        obj.allow(user, "write")
    with code.ContextManager(user, _writer):
        obj = lookup("thing")
        obj.owner = t_wizard
        obj.save()


@pytest.mark.django_db(transaction=True, reset_sequences=True)
def test_cant_change_location_unless_allowed_to_move(t_init: Object, t_wizard: Object):
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.ContextManager(t_wizard, _writer):
        obj = create("thing")
    user = Object.objects.get(name__iexact="player")
    with code.ContextManager(user, _writer):
        obj = lookup("thing")
        obj.location = user
        with pytest.raises(PermissionError) as excinfo:
            obj.save()
        assert str(excinfo.value) == f"#{user.pk} (Player) is not allowed write on #{obj.pk} (thing)"
    with code.ContextManager(t_wizard, _writer):
        obj = lookup("thing")
        obj.allow(user, "move")
        obj.allow(user, "write")
    with code.ContextManager(user, _writer):
        obj = lookup("thing")
        obj.location = user
        obj.save()


@pytest.mark.django_db(transaction=True, reset_sequences=True)
def test_change_location_calls_enterfunc(t_init: Object, t_wizard: Object, caplog: pytest.LogCaptureFixture):
    printed = []

    def _writer(msg):
        printed.append(msg)

    with caplog.at_level(logging.INFO, "moo.core.tasks.background"):
        with code.ContextManager(t_wizard, _writer):
            containers = lookup("container class")
            box = create("box", parents=[containers])
            box.add_verb("enterfunc", code="print(args[0])")
            thing = create("thing")
            thing.location = box
            thing.save()
    assert f"#{thing.pk} (thing)\n" in caplog.text


@pytest.mark.django_db(transaction=True, reset_sequences=True)
def test_change_location_calls_exitfunc(t_init: Object, t_wizard: Object, caplog: pytest.LogCaptureFixture):
    printed = []

    def _writer(msg):
        printed.append(msg)

    with caplog.at_level(logging.INFO, "moo.core.tasks.background"):
        with code.ContextManager(t_wizard, _writer):
            containers = lookup("container class")
            box = create("box", parents=[containers])
            box.add_verb("exitfunc", code="print(args[0])")
            thing = create("thing", location=box)
            assert thing.location == box
        with code.ContextManager(t_wizard, _writer):
            thing = lookup("thing")
            thing.location = t_wizard.location
            thing.save()
            assert thing.location == t_wizard.location
    assert f"#{thing.pk} (thing)\n" in caplog.text


@pytest.mark.django_db(transaction=True, reset_sequences=True)
def test_change_location_calls_accept(t_init: Object, t_wizard: Object):
    printed = []

    def _writer(msg):
        printed.append(msg)

    user = Object.objects.get(name__iexact="player")
    with code.ContextManager(user, _writer):
        box = create("box")
        box.add_verb("accept", code="return False")
        with pytest.raises(PermissionError) as excinfo:
            thing = create("thing", location=box)
        thing = lookup("thing")
        assert str(excinfo.value) == f"#{box.pk} (box) did not accept #{thing.pk} (thing)"


@pytest.mark.django_db(transaction=True, reset_sequences=True)
def test_perm_cache_avoids_repeated_queries(t_init: Object, t_wizard: Object):
    thing = Object.objects.create(name="thing", owner=t_wizard)
    with code.ContextManager(t_wizard, lambda m: None):
        # First call populates the cache.
        assert t_wizard.is_allowed("read", thing)
        queries_after_first = len(connection.queries)
        # Second call for the same (caller, permission, subject) should hit the cache.
        assert t_wizard.is_allowed("read", thing)
        assert len(connection.queries) == queries_after_first


@pytest.mark.django_db(transaction=True, reset_sequences=True)
def test_perm_cache_invalidated_by_deny(t_init: Object, t_wizard: Object):
    user = Object.objects.get(name__iexact="player")
    thing = Object.objects.create(name="thing", owner=user)
    with code.ContextManager(t_wizard, lambda m: None):
        # Populate cache with True for "read".
        assert user.is_allowed("read", thing)
        # Add a deny rule — cache should be evicted.
        thing.deny(user, "read")
        # Now the result must reflect the deny rule, not the stale cache.
        assert not user.is_allowed("read", thing)


@pytest.mark.django_db(transaction=True, reset_sequences=True)
def test_change_location_checks_recursion(t_init: Object, t_wizard: Object):
    printed = []

    def _writer(msg):
        printed.append(msg)

    with code.ContextManager(t_wizard, _writer):
        containers = lookup("container class")
        box = create("box", parents=[containers])
        envelope = create("envelope", parents=[containers], location=box)
        with pytest.raises(exceptions.RecursiveError) as excinfo:
            box.location = envelope
            box.save()
        assert str(excinfo.value) == f"#{box.pk} (box) already contains #{envelope.pk} (envelope)"



# Generated by Django 5.0.3 on 2024-05-06 16:18

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("core", "0009_alter_access_unique_together"),
    ]

    operations = [
        migrations.AddField(
            model_name="object",
            name="obvious",
            field=models.BooleanField(default=True),
        ),
    ]



# -*- coding: utf-8 -*-
from .base import *  # pylint: disable=wildcard-import

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "django-insecure-061+p62f39ohlfrgu&)%1lxo%%#_-$rc5l_zsrlx6jqy)sw(=r"

DATABASES = {
    "default": {"ENGINE": "django.db.backends.sqlite3", "NAME": BASE_DIR / "db.sqlite3", "TEST": {"NAME": ":memory:"}}
}

# All tests that move objects around while having enterfuncs and exitfuncs
# need to run tasks eagerly to ensure that the enterfuncs and exitfuncs are
# executed within the same test transaction.
CACHES = {
    "default": {
        "BACKEND": "django.core.cache.backends.locmem.LocMemCache"
    }
}

# Disable cross-session property cache in tests: the in-process LocMemCache
# does not reset between test cases, which would poison subsequent tests
# when PKs are reused after sequence resets.
MOO_PROP_CACHE_TTL = 0



#!moo verb open --on $container --dspec this

# pylint: disable=return-outside-function,undefined-variable

"""
This verb will open the container and allow objects to be put into it (via the put verb).
This verb sets the property `opened` to `True`.
"""

from moo.core import context

if this.is_open():
    print("Container is already open.")
elif this.is_openable_by(context.player):
    this.set_opened(True)
    print("You open the container.")



#!moo verb tell_lines --on $root_class

# pylint: disable=return-outside-function,undefined-variable

"""
This outputs out the list of strings strings to the object, using the tell verb for this object. Each string in strings
is output on a separate line.
"""

strings = args[0]

for arg in strings:
    this.tell(arg)



# -*- coding: utf-8 -*-
import asyncio

from django.core.management.base import BaseCommand

from ...server import server


class Command(BaseCommand):
    help = "Run the moo SSH server."

    def handle(self, *args, **options):
        import logging

        logging.info("Starting shell server...")
        asyncio.run(server())
